! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
!
! Module: esas_mod
!
! Authors:  Steve Snowden - NASA/GSFC
!           Kip Kuntz - NASA/GSFC/JHU
!           Brendan Perry - NASA/GSFC F77-f95 conversion and SAS compliance
! Comments: subroutine library for use by espfilt and new ESAS tasks (SAS-19)
!
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
! *** BP rewrite of Snowden esas "subs.f" routines ***
!     for SAS/F90/gfortran compliance
!
module esas_mod

  use dal
  use param 
  use errorHandling
  use gofio_mod
  use gofutils_mod
  use gofmatrix_mod
  use gofcoords_mod

  implicit none

contains

  subroutine aug_get_exp (qdp_F,QPB_S,ratein,hardin,revlin,c_dst,c_tim,&
       doplot,didplot,aug_spec,aug_expo,mosnum,ccdN,chans,obsid,expid)
    !
    ! called by ESAS routine(s):
    !  mosback (old mos_back)
    !
    ! Inputs:
    !   QPB_S: quiescent particle background mos1-qpb.fits set already opened
    !   ratein: R*4 
    !   hardin: R*4
    !   revlin: I*4
    !   c_dst(3) (/0.01,1.,0.1/) hardcoded in mosback R*4
    !   c_tim    1E06 hardcoded in mosback            R*4
    !   doplot   usually 'withqdp' logical  L*1
    !   didplot  second pass, don't re-open  L*1
    !   aug_spec(chans) OUT augmented spectra R*4
    !   aug_expo        OUT augmented exposure R*4
    !   mosnum          1 or 2 I*4 
    !   ccdN            2-7    I*4
    !   chans           2400 for mosback, variable I*4
    !   obsid           1234567890 C*10
    !   prefix          1S002      C*5
    !
    type(DataSetT), intent(inout) :: QPB_S
    type(TableT)   :: QPB_T
    type(ColumnT)  :: obsid_C, expid_C, object_C, date_C, back_C, &
         revl_C, filter_C, ctime_C, expo_C, rate_C, submode_C, &
         urate_C, hard_C, uhard_C, spectrum_C, ttime_C, key_C

    character(len=15) :: ecode
    character(len=1) :: cccd
    character(len=20), allocatable, dimension(:) :: obid, objt
    character(len=20) :: extname
    character(len=24) :: title
    character(len=25), dimension(:), allocatable :: filt
    character(len=*), intent(in) :: qdp_F, expid, obsid
    integer, intent(in) :: chans, ccdN, mosnum, revlin

    integer(kind=int32), dimension(:,:), pointer :: spec_ptr
    integer(kind=int32), dimension(:), pointer :: revl_ptr
    integer :: i, ii, lunqdp, ikeep, iodd, itd, nchan, &
         ncols, nobs, n, qpbRows, status = 0
    integer, allocatable, dimension(:) :: revl, ord
    logical, intent(in) :: doplot
    logical, intent(inout) :: didplot
    logical :: warndie = .TRUE.
    real(kind=single), intent(in) :: c_tim, c_dst(3), ratein, hardin
    real(kind=single), intent(inout) :: aug_expo
    real(kind=single) :: ave_rate, cum_exp, &
         d1, d2, hardmax, hardmin, muhard, &
         murate, ratemax, ratemin, td, temphard, temp, tr
    ! bhp: old was 2400 channel spectra and 7500 QPB rows
    !      now, chans spectra rows allocatable
    real(kind=single) :: aug_spec(chans) ! 2400
    real(kind=single), allocatable, dimension(:,:) :: spec ! (2400,7500)
    real(kind=single), allocatable, dimension(:) :: ddist, dist,&
         energ, expo, hard, hdist, rate, rdist, tdist, time, &
         uhard, urate
    real(kind=double), allocatable, dimension(:) :: date
    ! old 7500: date,ddist,dist,expo,hard,hdist,rate,rdist,tdist,
    !           time,uhard,urate
    ! old 2400: energ
    !    
    real(kind=double), dimension(:), pointer :: expo_ptr, rate_ptr, &
         urate_ptr, hard_ptr, uhard_ptr
    !
    ! qdpfile was augfile
    ! now input from calling routine
    ! qdp_F = "mos" // adjustl(trim(prefix)) // "-aug.qdp"
    ! create title for qdp plot
    title = '"' // adjustl(trim(obsid)) // ' -- MOS' // &
         adjustl(trim(expid)) // '"'
    !
    !  If selected, make a plot, first set up the plot
    !
    if(doplot .and. .not. didplot) then
       write(errstr,*)"Writing out QDP plot file: ",trim(qdp_F)
       call message(AppMsg,Always)
       !
       ! open the qdp file for writing
       !
       call open_ascii_file(qdp_F,lunqdp)
       write(lunqdp,*) ' viewport 0.2 0.2 0.9 0.8'
       write(lunqdp,*) ' la t ', title
       write(lunqdp,*) ' read serr 1'
       write(lunqdp,*) ' read serr 2'
       write(lunqdp,*) ' read serr 3'
       write(lunqdp,*) ' read serr 4'
       write(lunqdp,*) ' read serr 5'
       write(lunqdp,*) ' read serr 6'
       write(lunqdp,*) ' read serr 7'
       write(lunqdp,*) ' read serr 8'
       write(lunqdp,*) ' read serr 9'
       write(lunqdp,*) ' read serr 10'
       write(lunqdp,*) ' read serr 11'
       write(lunqdp,*) ' read serr 12'
       write(lunqdp,*) ' read serr 13'
       write(lunqdp,*) ' read serr 14'
       write(lunqdp,*) ' read serr 15'
       write(lunqdp,*) ' read serr 16'
       write(lunqdp,*) ' read serr 17'
       write(lunqdp,*) ' read serr 18'
       write(lunqdp,*) ' read serr 19'
       write(lunqdp,*) ' font roman'
       write(lunqdp,*) ' cs 1.5'
       write(lunqdp,*) ' la f'
       write(lunqdp,*) ' col 2 on 2'
       write(lunqdp,*) ' col 1 on 3'
       write(lunqdp,*) ' col 3 on 4'
       write(lunqdp,*) ' col 2 on 5'
       write(lunqdp,*) ' col 1 on 6'
       write(lunqdp,*) ' col 3 on 7'
       write(lunqdp,*) ' col 2 on 8'
       write(lunqdp,*) ' col 1 on 9'
       write(lunqdp,*) ' col 3 on 10'
       write(lunqdp,*) ' col 2 on 11'
       write(lunqdp,*) ' col 1 on 12'
       write(lunqdp,*) ' col 3 on 13'
       write(lunqdp,*) ' col 2 on 14'
       write(lunqdp,*) ' col 1 on 15'
       write(lunqdp,*) ' col 3 on 16'
       write(lunqdp,*) ' col 2 on 17'
       write(lunqdp,*) ' col 1 on 18'
       write(lunqdp,*) ' col 3 on 19'
       write(lunqdp,*) ' mark 1 on 2'
       write(lunqdp,*) ' mark 1 on 3'
       write(lunqdp,*) ' mark 1 on 4'
       write(lunqdp,*) ' mark 1 on 5'
       write(lunqdp,*) ' mark 1 on 6'
       write(lunqdp,*) ' mark 1 on 7'
       write(lunqdp,*) ' mark 1 on 8'
       write(lunqdp,*) ' mark 1 on 9'
       write(lunqdp,*) ' mark 1 on 10'
       write(lunqdp,*) ' mark 1 on 11'
       write(lunqdp,*) ' mark 1 on 12'
       write(lunqdp,*) ' mark 1 on 13'
       write(lunqdp,*) ' mark 1 on 14'
       write(lunqdp,*) ' mark 1 on 15'
       write(lunqdp,*) ' mark 1 on 16'
       write(lunqdp,*) ' mark 1 on 17'
       write(lunqdp,*) ' mark 1 on 18'
       write(lunqdp,*) ' mark 1 on 19'
       write(lunqdp,*) ' line off 2'
       write(lunqdp,*) ' line off 3'
       write(lunqdp,*) ' line off 4'
       write(lunqdp,*) ' line off 5'
       write(lunqdp,*) ' line off 6'
       write(lunqdp,*) ' line off 7'
       write(lunqdp,*) ' line off 8'
       write(lunqdp,*) ' line off 9'
       write(lunqdp,*) ' line off 10'
       write(lunqdp,*) ' line off 11'
       write(lunqdp,*) ' line off 12'
       write(lunqdp,*) ' line off 13'
       write(lunqdp,*) ' line off 14'
       write(lunqdp,*) ' line off 15'
       write(lunqdp,*) ' line off 16'
       write(lunqdp,*) ' line off 17'
       write(lunqdp,*) ' line off 18'
       write(lunqdp,*) ' line off 19'
       write(lunqdp,*) ' win 1'
       write(lunqdp,*) ' yplot 1 2 3 4'
       write(lunqdp,*) ' loc 0.05 0.48 0.39 1.0'
       write(lunqdp,*) ' la 11 p 0.06 9.0 just left "CCD #2"'
       write(lunqdp,*) ' r x1 0.0 0.1'
       write(lunqdp,*) ' r y1 0.0 10.0'
       write(lunqdp,*) ' la y Hardness'
       write(lunqdp,*) ' win 2'
       write(lunqdp,*) ' yplot 1 5 6 7'
       write(lunqdp,*) ' loc 0.35 0.48 0.69 1.0'
       write(lunqdp,*) ' la 12 p 0.06 9.0 just left "CCD #3"'
       write(lunqdp,*) ' r x2 0.0 0.1'
       write(lunqdp,*) ' r y2 0.0 10.0'
       write(lunqdp,*) ' win 3'
       write(lunqdp,*) ' yplot 1 8 9 10'
       write(lunqdp,*) ' loc 0.65 0.48 0.99 1.0'
       write(lunqdp,*) ' la 13 p 0.06 9.0 just left "CCD #4"'
       write(lunqdp,*) ' r x3 0.0 0.1'
       write(lunqdp,*) ' r y3 0.0 10.0'
       write(lunqdp,*) ' win 4'
       write(lunqdp,*) ' yplot 1 11 12 13'
       write(lunqdp,*) ' loc 0.05 0.05 0.39 0.57'
       write(lunqdp,*) ' la 14 p 0.06 9.0 just left "CCD #5"'
       write(lunqdp,*) ' r x4 0.0 0.1'
       write(lunqdp,*) ' r y4 0.0 10.0'
       write(lunqdp,*) ' la y Hardness'
       write(lunqdp,*) ' la x Rate'
       write(lunqdp,*) ' win 5'
       write(lunqdp,*) ' yplot 1 14 15 16'
       write(lunqdp,*) ' loc 0.35 0.05 0.69 0.57'
       write(lunqdp,*) ' la 15 p 0.06 9.0 just left "CCD #6"'
       write(lunqdp,*) ' r x5 0.0 0.1'
       write(lunqdp,*) ' r y5 0.0 10.0'
       write(lunqdp,*) ' la x Rate'
       write(lunqdp,*) ' win 6'
       write(lunqdp,*) ' yplot 1 17 18 19'
       write(lunqdp,*) ' loc 0.65 0.05 0.99 0.57'
       write(lunqdp,*) ' la 16 p 0.06 9.0 just left "CCD #7"'
       write(lunqdp,*) ' r x6 0.0 0.1'
       write(lunqdp,*) ' r y6 0.0 10.0'
       write(lunqdp,*) ' la x Rate'
    endif
    !
    !  Figure the correct extension name by chip number
    !
    write(cccd,'(i1)')ccdN
    extname = "QUIES_CHIP" // trim(cccd)
    ecode = "noQPB" // cccd // "exten"
    write(errstr,*)"Checking QPB file: for extension: ",trim(extname)
    call message(AppMsg,Always)
    if (.not.hasBlock(QPB_S,trim(extname))) then
       !
       ! here is the old 'AUG_GET_EXP POSITION 04' error
       ! ['Could not find the spectrum extension'] from the f77
       !
       write(errstr,*)&
            "CAL QPB file does not contain extension : ",trim(extname)
       call error(trim(adjustl(ecode)),trim(errstr))
    else
       QPB_T = table(QPB_S,trim(extname)) ! read the whole extension
       qpbrows = numberOfRows(QPB_T)      ! get number of rows
       write(errstr,*)"CAL QPB file has ",qpbrows, "rows"
       call message(AppMsg,Always)
    endif
    !
    !*** now read the colums in mos*-qpb.fits file (BHP) ***
    !*** Columns = obsid, name, object, date, revolut, filter, ontime,
    !*** expo, rate, rate_err, hard, hard_err, spectrum
    !*** most aren't actually used in calculations here (BHP)
    !
    write(errstr,*)"Reading relevant columns from QPB SPECTRUM extension"
    call message(AppMsg,Noisy)
    if (.not.hasColumn(QPB_T,"OBSID")) then
       call error("badColumn","Input file incorrect - no OBSID column")
    else
       object_C = column(QPB_T,"OBSID",READ)
    endif
    if (.not.hasColumn(QPB_T,"EXP_ID")) then
       call error("badColumn","Input file incorrect - no EXP_ID column")
    else
       expid_C = column(QPB_T,"OBSID",READ)
    endif
    if (.not.hasColumn(QPB_T,"DATE")) then
       call error("badColumn","Input file incorrect - no DATE column")
    else
       date_C = column(QPB_T,"DATE",READ)
    endif
    if (.not.hasColumn(QPB_T,"REVOLUT")) then
       call error("badColumn","Input file incorrect - no REVOLUT column")
    else
       revl_C = column(QPB_T,"REVOLUT",READ)
       allocate(revl(qpbRows))
       revl_ptr => int32Data(revl_C)
       revl = revl_ptr
    endif
    if (.not.hasColumn(QPB_T,"SUB_MODE")) then
       call error("badColumn","Input file incorrect - no SUB_MODE column")
    else
       submode_C = column(QPB_T,"SUB_MODE",READ)
    endif
    if (.not.hasColumn(QPB_T,"FILTER")) then
       call error("badColumn","Input file incorrect - no FILTER column")
    else
       filter_C = column(QPB_T,"FILTER",READ)
       allocate(filt(qpbRows))
       do i=1,qpbrows
          filt(i) = stringCell(filter_C,i-1)  ! fill chans array with keys
       end do
    endif
    if (.not.hasColumn(QPB_T,"TOTAL_TIME")) then
       call error("badColumn","Input file incorrect - no TOTAL_TIME column")
    else
       ttime_C = column(QPB_T,"TOTAL_TIME",READ)
    endif
    if (.not.hasColumn(QPB_T,"CLEAN_TIME")) then
       call error("badColumn","Input file incorrect - no CLEAN_TIME column")
    else
       ctime_C = column(QPB_T,"CLEAN_TIME",READ)
    endif
    if (.not.hasColumn(QPB_T,"KEY")) then
       call error("badColumn","Input file incorrect - no KEY column")
    else
       key_C = column(QPB_T,"KEY",READ)
    endif
    if (.not.hasColumn(QPB_T,"EXPO")) then
       call error("badColumn","Input file incorrect - no EXPO column")
    else
       expo_C = column(QPB_T,"EXPO",READ)
       expo_ptr => Real64Data(expo_C)
       allocate(expo(qpbrows))
       expo = expo_ptr
    endif
    if (.not.hasColumn(QPB_T,"BACKSCAL")) then
       call error("badColumn","Input file incorrect - no BACKSCAL column")
    else
       back_C = column(QPB_T,"BACKSCAL",READ)
    endif
    if (.not.hasColumn(QPB_T,"RATE")) then
       call error("badColumn","Input file incorrect - no RATE column")
    else
       rate_C = column(QPB_T,"RATE",READ)
       rate_ptr => Real64Data(rate_C)
       allocate(rate(qpbrows))
       rate = rate_ptr
    endif
    if (.not.hasColumn(QPB_T,"RATE_ERR")) then
       call error("badColumn","Input file incorrect - no RATE_ERR column")
    else
       urate_C = column(QPB_T,"RATE_ERR",READ)
       urate_ptr => Real64Data(urate_C)
       allocate(urate(qpbrows))
       urate = urate_ptr
    endif
    if (.not.hasColumn(QPB_T,"HARD")) then
       call error("badColumn","Input file incorrect - no HARD column")
    else
       hard_C = column(QPB_T,"HARD",READ)
       hard_ptr => Real64Data(hard_C)
       allocate(hard(qpbrows))
       hard = hard_ptr
    endif
    if (.not.hasColumn(QPB_T,"HARD_ERR")) then
       call error("badColumn","Input file incorrect - no HARD_ERR column")
    else
       uhard_C = column(QPB_T,"HARD_ERR",READ)
       uhard_ptr => Real64Data(uhard_C)
       allocate(uhard(qpbrows))
       uhard = uhard_ptr
    endif
    if (.not.hasColumn(QPB_T,"SPECTRUM")) then
       call error("badColumn","Input file incorrect - no SPECTRUM column")
    else
       ! don't need tspec with dal call
       spectrum_C = column(QPB_T,"SPECTRUM",READ)
       spec_ptr => int32VectorData(spectrum_C)
       allocate(spec(chans,qpbrows))
       spec = spec_ptr(0:chans-1,0:qpbrows-1)
    endif
    !
    !  Do some cleaning, and selecting on CCD state
    !
    !  first determine if the observation CCD is in an odd state
    !
    iodd = 0
    if(mosnum == 1) then
       if (ccdN == 4) then
          tr = 100.0 * ratein
          if(tr < 5.2) then
             temphard = 0.96
             !elseif(tr < 5.2) then
             !temphard = 0.423*tr - 1.24
          else
             temphard = 0.423*tr - 1.24
          endif
          if(hardin < temphard) iodd = 1
       elseif (ccdN == 5) then
          tr = 100.0 * ratein
          if(tr < 7.0) then
             temphard = 1.54
          elseif (tr < 9.0) then
             temphard = 0.53*tr - 2.17
          else
             temphard = 2.60
          endif
          if(hardin < temphard) iodd = 1
       else
          if(hardin < 1.5) iodd = 1
       endif
    else
       if (ccdN == 2) then
          if(hardin < 1.1) iodd = 1
       elseif (ccdN == 5) then
          tr = 100.0 * ratein
          if(tr < 10.0) then
             temphard = 0.15*tr
          elseif (tr < 16.67) then
             temphard = 0.15*tr
          else
             temphard = 2.50
          endif
          if(hardin < temphard) iodd = 1
       else
          if(hardin < 1.5) iodd = 1
       endif
    endif
    !
    !  Set the normal filtering limits
    !
    d1 = -9.0
    d2 = 0.0
    hardmin = 0.2
    hardmax = 15.0 ! hardcoded or parameters??? BHP
    ratemin = 0.01
    ratemax = 0.1
    !
    !  Sort through the archived observation data
    !
    nobs = 0
    do i=1,qpbrows
       if((expo(i) >= 1000.0) .and. (rate(i) > ratemin) .and.&
            (rate(i) < ratemax) .and.&
            (hard(i) > hardmin) .and.&
            (hard(i) < hardmax) .and.&
            (filt(i)(:6) .ne. 'CalClo')) then
          !
          !  Write out all the data for plotting if selected
          !
          if(doplot) then
             select case (ccdN)
             case (2) 
                write(lunqdp,2120) rate(i), urate(i),&
                     hard(i), uhard(i), d1, d2,&
                     d1, d2, d1, d2, d1, d2, d1, d2,&
                     d1, d2, d1, d2, d1, d2, d1, d2,&
                     d1, d2, d1, d2
2120            format(1x,2f10.6,2f6.2,22f4.0)
             case (3)
                write(lunqdp,2130) rate(i), urate(i),&
                     d1, d2, d1, d2,&
                     hard(i), uhard(i), d1, d2,&
                     d1, d2, d1, d2,&
                     d1, d2, d1, d2, d1, d2, d1, d2,&
                     d1, d2, d1, d2
2130            format(1x,2f10.6,4f4.0,2f6.2,18f4.0)
             case (4)
                write(lunqdp,2140) rate(i), urate(i),&
                     d1, d2, d1, d2, d1, d2, d1, d2,&
                     hard(i), uhard(i), d1, d2,&
                     d1, d2, d1, d2, d1, d2, d1, d2,&
                     d1, d2, d1, d2
2140            format(1x,2f10.6,8f4.0,2f6.2,14f4.0)
             case (5)
                write(lunqdp,2150) rate(i), urate(i),&
                     d1, d2, d1, d2, d1, d2, d1, d2,&
                     d1, d2, d1, d2,&
                     hard(i), uhard(i), d1, d2,&
                     d1, d2, d1, d2,&
                     d1, d2, d1, d2
2150            format(1x,2f10.6,12f4.0,2f6.2,10f4.0)
             case (6)
                write(lunqdp,2160) rate(i), urate(i),&
                     d1, d2, d1, d2, d1, d2, d1, d2,&
                     d1, d2, d1, d2, d1, d2, d1, d2,&
                     hard(i), uhard(i), d1, d2,&
                     d1, d2, d1, d2
2160            format(1x,2f10.6,16f4.0,2f6.2,6f4.0)
             case (7)
                write(lunqdp,2170) rate(i), urate(i),&
                     d1, d2, d1, d2, d1, d2, d1, d2,&
                     d1, d2, d1, d2, d1, d2, d1, d2,&
                     d1, d2, d1, d2,&
                     hard(i), uhard(i), d1, d2
2170            format(1x,2f10.6,20f4.0,2f6.2,2f4.0)
             case default ! should NEVER happen! 
                call error ("noCCDN",&
                     "aug_get_exp: ccd is none of: [2-7]")
             end select
          endif
          !
          !  Check to see whether the observation CCD is in an odd state
          !  and select whether the data will be kept or not
          !
          ikeep = 1
          if(mosnum == 1) then
             if (ccdN == 4) then
                tr = 100.0 * ratein
                if(tr < 5.2) then
                   temphard = 0.96
                   !                elseif(tr < 5.2) then
                   !temphard = 0.423*tr - 1.24
                else
                   temphard = 0.423*tr - 1.24
                endif
                if(iodd == 1) then
                   if(hard(i) >= temphard) ikeep = 0
                else
                   if(hard(i) < temphard) ikeep = 0
                endif
             elseif (ccdN == 5) then
                tr = 100.0 * ratein
                if(tr < 7.0) then
                   temphard = 1.54
                elseif (tr < 9.0) then
                   temphard = 0.53*tr - 2.17
                else
                   temphard = 2.60
                endif
                if(iodd == 1) then
                   if(hard(i) >= temphard) ikeep = 0
                else
                   if(hard(i) < temphard) ikeep = 0
                endif
             else
                if(hard(i) < 1.5) then
                   ikeep = 0
                endif
             endif
          endif
          if(mosnum == 2) then
             if(ccdN == 2) then
                if(iodd == 1) then
                   if(hard(i) >= 1.1) ikeep = 0
                else
                   if(hard(i) < 1.1) ikeep = 0
                endif
             elseif (ccdN == 5) then
                tr = 100.0 * ratein
                if(tr < 10.0) then
                   temphard = 0.15*tr
                elseif (tr < 16.67) then
                   temphard = 0.15*tr
                else
                   temphard = 2.50
                endif
                if(iodd == 1) then
                   if(hard(i)  >= temphard) ikeep = 0
                else
                   if(hard(i)  < temphard) ikeep = 0
                endif
             else
                if (iodd == 1) then
                   if(hard(i)  >= 1.5) ikeep = 0
                else
                   if(hard(i)  < 1.5) ikeep = 0
                endif
             endif
          endif
          if(adjustl(trim(filt(i)(:6))) == 'CalClo') ikeep = 0
          !
          !  For all where ikeep=1 transfer the information
          !
          if(ikeep == 1) then
             nobs = nobs + 1
             revl(nobs) = revl(i)
             expo(nobs) = expo(i)
             rate(nobs) = rate(i)
             urate(nobs) = urate(i)
             hard(nobs) = hard(i)
             uhard(nobs) = uhard(i)
             do ii=1,chans
                spec(ii,nobs) = spec(ii,i)
             enddo
          endif
       endif
    enddo
    !
    !  Set the energies of the bins
    !
    allocate(energ(chans))
    do i=1,chans
       energ(i) = 0.005*(i - 1)
    enddo
    !
    !  Get some temporary values
    !
    murate = 0.0
    muhard = 0.0
    do i=1,nobs
       murate = murate + urate(i)
       muhard = muhard + uhard(i)
    enddo
    murate = murate/nobs
    muhard = muhard/nobs
    !
    write(cccd,'(i1)')ccdN
    write(errstr,'(A,f12.5,A,f12.5)')'Rate: ',ratein,' Rate error: ',murate
    call message(AppMsg,Sparse)
    write(errstr,'(A,f12.5,A,f12.5)')'Hard: ',hardin,' Hard error: ',muhard
    call message(AppMsg,Sparse)
    !
    !  Find the closest c_tim secs
    !
    allocate(dist(qpbRows),ddist(qpbRows))
    allocate(hdist(qpbRows),rdist(qpbRows))
    allocate(tdist(qpbRows),ord(qpbRows))
    !
    if(c_tim == 0.0) then
       print *, 'Error: no exposure length provided'
       call error ("NoExpoLen","No exposure length provided")
       !       return ***return gracefully after that!???***
    else
       if((c_dst(1) == 0.0) .and. (c_dst(2) == 0.0) .and.&
            (c_dst(3) == 0.0)) then
          call error("NoDistScal","No distance scaling provided")
          !          return ***ditto
       else
          do i=1,nobs
             if(c_dst(1) > 0.0) then
                rdist(i) = (rate(i)-ratein)/c_dst(1)
             else
                rdist(i) = rate(i)*0.0
             endif
             if(c_dst(2) > 0.0) then
                hdist(i) = (hard(i)-hardin)/c_dst(2)
             else
                hdist(i) = hard(i)*0.0
             endif
             if(c_dst(3) > 0.0) then
                ddist(i) = (revl(i)-revlin)/c_dst(3)
             else
                ddist(i) = revl(i)*0.0
             endif
             dist(i) = sqrt(rdist(i)**2 + hdist(i)**2 + &
                  ddist(i)**2)
          enddo
       endif
       do i=1,nobs
          ord(i) = i
          tdist(i) = dist(i)
       enddo
       do ii=nobs-1,1,-1
          do i=1,ii
             if(tdist(i) > tdist(i+1)) then
                td = tdist(i)
                tdist(i) = tdist(i+1)
                tdist(i+1) = td
                itd = ord(i)
                ord(i) = ord(i+1)
                ord(i+1) = itd
             endif
          enddo
       enddo
       cum_exp = 0.0
       n = 0
       ave_rate = 0.0
       do while ((cum_exp < c_tim) .and. (n < nobs))
          n = n + 1
          cum_exp = cum_exp + expo(ord(n))
          ave_rate = ave_rate + rate(ord(n))*expo(ord(n))
          if(doplot) then
             if(ccdN == 2) then
                write(lunqdp,2125) rate(ord(n)), urate(ord(n)),&
                     d1, d2, hard(ord(n)), uhard(ord(n)),&
                     d1, d2, d1, d2, d1, d2, d1, d2,&
                     d1, d2, d1, d2, d1, d2, d1, d2,&
                     d1, d2, d1, d2
2125            format(1x,2f10.6,2f4.0,2f6.2,20f4.0)
             elseif(ccdN == 3) then
                write(lunqdp,2135) rate(ord(n)), urate(ord(n)),&
                     d1, d2, d1, d2,&
                     d1, d2, hard(ord(n)), uhard(ord(n)),&
                     d1, d2, d1, d2,&
                     d1, d2, d1, d2, d1, d2, d1, d2,&
                     d1, d2, d1, d2
2135            format(1x,2f10.6,6f4.0,2f6.2,16f4.0)
             elseif(ccdN == 4) then
                write(lunqdp,2145) rate(ord(n)), urate(ord(n)),&
                     d1, d2, d1, d2, d1, d2, d1, d2,&
                     d1, d2, hard(ord(n)), uhard(ord(n)),&
                     d1, d2, d1, d2, d1, d2, d1, d2,&
                     d1, d2, d1, d2
2145            format(1x,2f10.6,10f4.0,2f6.2,12f4.0)
             elseif(ccdN == 5) then
                write(lunqdp,2155) rate(ord(n)), urate(ord(n)),&
                     d1, d2, d1, d2, d1, d2, d1, d2,&
                     d1, d2, d1, d2,&
                     d1, d2, hard(ord(n)), uhard(ord(n)),&
                     d1, d2, d1, d2,&
                     d1, d2, d1, d2
2155            format(1x,2f10.6,14f4.0,2f6.2,8f4.0)
             elseif(ccdN == 6) then
                write(lunqdp,2165) rate(ord(n)), urate(ord(n)),&
                     d1, d2, d1, d2, d1, d2, d1, d2,&
                     d1, d2, d1, d2, d1, d2, d1, d2,&
                     d1, d2, hard(ord(n)), uhard(ord(n)),&
                     d1, d2, d1, d2
2165            format(1x,2f10.6,18f4.0,2f6.2,4f4.0)
             elseif(ccdN == 7) then
                write(lunqdp,2175) rate(ord(n)), urate(ord(n)),&
                     d1, d2, d1, d2, d1, d2, d1, d2,&
                     d1, d2, d1, d2, d1, d2, d1, d2,&
                     d1, d2, d1, d2,&
                     d1, d2, hard(ord(n)), uhard(ord(n))
2175            format(1x,2f10.6,22f4.0,2f6.2)
             endif
          endif
       enddo

       ave_rate = ave_rate/cum_exp
       write(errstr,4010) 'Exposure: ', cum_exp
       call message(AppMsg,Always)
4010   format(A,f10.2)
       write(errstr,*) 'Exposure weighted count rate: ', ave_rate
       call message(AppMsg,Always)
       write(errstr,4020)  'Observation count rate: ', ratein, &
            ' and ratio: ', ave_rate/ratein
       call message(AppMsg,Always)
4020   format(A,f12.4,A,f10.6)
    endif

    !
    !*** close the input fits files the DAL way:
    !
    call release(QPB_T)
    !
    !  Actually put together the spectrum
    !
    do ii=1,chans
       aug_spec(ii) = 0.0
    enddo
    aug_expo = 0.
    do i=1,n
       do ii=1,chans
          aug_spec(ii) = aug_spec(ii) + spec(ii,ord(i))
       enddo
       aug_expo = aug_expo + expo(ord(i))
    enddo
    aug_expo = aug_expo*ave_rate/ratein

    return

  end subroutine aug_get_exp
  !
  ! same as above, but for PN
  !
  subroutine augment_pn (QPB_F,ratein,hardin,revlin,c_dst,c_tim,&
       doplot,aug_spec,aug_expo,quadN,chans,obsid,prefix,submodein,&
       plc_lo,plc_hi)
    !
    ! called by ESAS routine(s):
    !  pnback (old pn_back)
    !
    ! Inputs:
    !   QPB_F: quiescent particle background CAL file pn-qpb.fits
    !   ratein: R*4 these come
    !   hardin: R*4 in from pnback
    !   revlin: I*4 call
    !   c_dst(3) (/0.01,1.,0.1/) from pnback R*4
    !   c_tim    1E06 in pnback             R*4
    !   doplot   usually 'withqdp' logical  L*1
    !   aug_spec(chans) OUT augmented spectra R*4
    !   aug_expo        OUT augmented exposure R*4
    !   quadN           1-4    I*4
    !   chans           4096 for pnback, variable I*4
    !   obsid           1234567890 C*10
    !   prefix          S002      C*4
    !
    type(DataSetT) :: QPB_S
    type(TableT)   :: QPB_T
    type(ColumnT)  :: expo_C, rate_C, urate_C, hard2_C, uhard2_C, &
         spec_C, specoo_C, key_C, revl_C, submode_C, ttime_C, back_C
    character(len=*), intent(in) :: QPB_F
    character(len=*), intent(in) :: prefix, submodein
    character(len=1), allocatable, dimension(:) :: key
    character(len=10) :: obsid
    character(len=20), allocatable, dimension(:) :: submode
    character(len=20) :: extname, ecode
    character(len=24) :: title
    character(len=25) :: smode
    character(len=256) :: qdpaug_F, qdpspec_F, qdprate_F, qdphard_F
    character(len=1024) :: expression
    integer, intent(in) :: chans, quadN
    integer(kind=int32), dimension(:), pointer :: revl_ptr, back_ptr
    integer(kind=int32), dimension(:,:), pointer :: spec_ptr, specoo_ptr 
    !    integer(kind=int32), pointer :: arrayDataInt32(:,:)
    !
    integer :: i, lunaug, lunspec, lunrate, lunhard, ii, revlin, itd, &
         qpbRows, plc_lo(2), plc_hi(2), nb1, nb2, nb3, nobs, nbr, ngr, ngc, n
    integer, allocatable, dimension(:) :: revl, ord
    logical, intent(in) :: doplot
    logical, allocatable, dimension(:) :: goodrows
    real(kind=single), parameter :: amscal = 1440000.0
    real(kind=single), intent(in) :: c_tim, ratein, hardin, c_dst(3)
    real(kind=single), intent(inout) :: aug_expo
    real(kind=single) :: ave_rate, cum_exp, d1, d2, &
         hardmax, hardmin, muhard, murate, ratemax, ratemin, td, &
         temp
    !
    ! bhp: old was 4096 channel spectra and 7500 QPB rows
    !      now, chans spectra and maxQPB rows allocatable
    !      only rate and hard2 actually used
    real(kind=single) :: te1, te2, teo1, teo2, tootsca, expmin, expmax, &
         oootsca
    real(kind=single) :: csc(4,4), aug_spec(chans) ! 4096
    real(kind=single), allocatable, dimension(:,:) :: spec, specoo
    real(kind=double), allocatable, dimension(:) :: ddist, dist, hdist, &
         hard2, uhard2, rate, urate, ttime, rdist, tdist, tspec, &
         tspecoo, energ, expo, back, ootsca
    ! old 7500: date,ddist,dist,expo,hard,hdist,rate,rdist,tdist,
    !           tspec,tspecoo,time,uhard,urate
    ! old 4096: energ
    !    
    real(kind=double), dimension(:), pointer :: expo_ptr, rate_ptr, urate_ptr,&
         hard2_ptr, uhard2_ptr, ttime_ptr
    !
    ! expression for selectlib to find all spectra matching 
    !
    
    !
    !  Set the normal filtering limits
    !
    d1 = -9.0
    d2 = 0.0
    hardmin = 3.0
    hardmax = 7.0
    ratemin = 1.0e-3
    ratemax = 3.0e-3
    expmin = 5000.0
    expmax = 40000.0
    ! 
    csc(1,1) = 3.75
    csc(2,1) = 5.07
    csc(3,1) = 5.05
    csc(4,1) = 4.00
    csc(1,1) = 1.75
    csc(2,2) = 1.75
    csc(3,2) = 1.72
    csc(4,2) = 2.16

    write(errstr,4001) ratein,hardin,revlin
    call message(AppMsg,Always)
4001 format('Count Rate, Hardness, and Revolution: ',f8.6,f8.3,i5)

    write(errstr,4002) c_dst
    call message(AppMsg,Always)
4002 format('Scaling Factors for Obs. Selection: ',f8.6,f8.3,f8.0)
!!! make this a call line from pnback, it's already been done there!!!
    if(adjustl(trim(submodein)) == 'PrimeFullWindow') then
       write(errstr,*)"PN mode: ",submodein 
       call message(AppMsg,Always)
       smode = submodein
       oootsca = 0.063
    elseif(adjustl(trim(submodein)) == 'PrimeFullWindowExtended3') then
       write(errstr,*)"PN mode: ",submodein 
       call message(AppMsg,Always)
       smode = 'PrimeFullWindowExten'
       oootsca = 0.0232
    elseif(adjustl(trim(submodein)) == 'PrimeFullWindowExtended5') then
       write(errstr,*)"PN mode: ",submodein
       call message(AppMsg,Always)
       smode = 'PrimeFullWindowExten'
       oootsca = 0.0163
    else
       write(errstr,*)"augment_pn: Bad submode: ",submodein
       call error("BadSubMode",trim(errstr))
    endif
    write(errstr,4005)oootsca    
    call message(AppMsg,Always)
4005 format('Observation OOT Scale Factor: ',f7.4)
    !
    !  If selected, make a plot, first set up the plot
    !
    title = '"' // adjustl(trim(obsid)) // ' -- PN' // &
         adjustl(trim(prefix)) // '"'
    if(doplot) then
       write(errstr,*)&
            "Writing four QDP plots: aug,-spec,-rev-rate,-rev-hard"
       call message(AppMsg,Verbose)
       !
       ! open the qdp file for writing
       !
       qdpaug_F = "pn" // adjustl(trim(prefix)) // "-aug.qdp"
       ! create title for qdp plot
       call open_ascii_file(qdpaug_F,lunaug)
       write(lunaug,*) ' viewport 0.2 0.2 0.9 0.8'
       write(lunaug,*) ' la t ',adjustl(trim(title))
       write(lunaug,*) ' read serr 1'
       write(lunaug,*) ' read serr 2'
       write(lunaug,*) ' read serr 3'
       write(lunaug,*) ' read serr 4'
       write(lunaug,*) ' read serr 5'
       write(lunaug,*) ' read serr 6'
       write(lunaug,*) ' read serr 7'
       write(lunaug,*) ' read serr 8'
       write(lunaug,*) ' read serr 9'
       write(lunaug,*) ' font roman'
       write(lunaug,*) ' cs 1.5'
       write(lunaug,*) ' la f'
       write(lunaug,*) ' col 2 on 2'
       write(lunaug,*) ' col 1 on 3'
       write(lunaug,*) ' col 2 on 4'
       write(lunaug,*) ' col 1 on 5'
       write(lunaug,*) ' col 2 on 6'
       write(lunaug,*) ' col 1 on 7'
       write(lunaug,*) ' col 2 on 8'
       write(lunaug,*) ' col 1 on 9'
       write(lunaug,*) ' mark 1 on 2'
       write(lunaug,*) ' mark 1 on 3'
       write(lunaug,*) ' mark 1 on 4'
       write(lunaug,*) ' mark 1 on 5'
       write(lunaug,*) ' mark 1 on 6'
       write(lunaug,*) ' mark 1 on 7'
       write(lunaug,*) ' mark 1 on 8'
       write(lunaug,*) ' mark 1 on 9'
       write(lunaug,*) ' line off 2'
       write(lunaug,*) ' line off 3'
       write(lunaug,*) ' line off 4'
       write(lunaug,*) ' line off 5'
       write(lunaug,*) ' line off 6'
       write(lunaug,*) ' line off 7'
       write(lunaug,*) ' line off 8'
       write(lunaug,*) ' line off 9'
       write(lunaug,*) ' win 1'
       write(lunaug,*) ' yplot 1 2 3'
       write(lunaug,*) ' loc 0.05 0.45 0.57 1.0'
       write(lunaug,*) ' la 11 p 0.00065 2.6 just left "Quad #1"'
       write(lunaug,*) ' r x1 0.0005 0.004'
       write(lunaug,*) ' r y1 2.0 8.0'
       write(lunaug,*) ' la y Hardness'
       write(lunaug,*) ' win 2'
       write(lunaug,*) ' yplot 1 4 5'
       write(lunaug,*) ' loc 0.47 0.45 0.99 1.0'
       write(lunaug,*) ' la 12 p 0.00065 2.6 just left "Quad #2"'
       write(lunaug,*) ' r x2 0.0005 0.004'
       write(lunaug,*) ' r y2 2.0 8.0'
       write(lunaug,*) ' win 3'
       write(lunaug,*) ' yplot 1 6 7'
       write(lunaug,*) ' loc 0.05 0.05 0.57 0.60'
       write(lunaug,*) ' la 13 p 0.00065 2.6 just left "Quad #3"'
       write(lunaug,*) ' r x3 0.0005 0.004'
       write(lunaug,*) ' r y3 2.0 8.0'
       write(lunaug,*) ' la y Hardness'
       write(lunaug,*) ' la x Rate'
       write(lunaug,*) ' win 4'
       write(lunaug,*) ' yplot 1 8 9'
       write(lunaug,*) ' loc 0.47 0.05 0.99 0.60'
       write(lunaug,*) ' la 14 p 0.00065 2.6 just left "Quad #4"'
       write(lunaug,*) ' r x4 0.0005 0.004'
       write(lunaug,*) ' r y4 2.0 8.0'
       write(lunaug,*) ' col off 1'
       write(lunaug,*) ' la x Rate'
       !
       ! open the qdp file for writing
       !
       qdpspec_F = "pn" // adjustl(trim(prefix)) // "-aug-spec.qdp"
       ! create title for qdp plot
       call open_ascii_file(qdpspec_F,lunspec)
       write(lunspec,*) ' viewport 0.2 0.2 0.9 0.8'
       write(lunspec,*) ' la t ', adjustl(trim(title))
       write(lunspec,*) ' font roman'
       write(lunspec,*) ' cs 1.5'
       write(lunspec,*) ' la f'
       write(lunspec,*) ' col 1 on 2'
       write(lunspec,*) ' col 1 on 3'
       write(lunspec,*) ' col 1 on 4'
       write(lunspec,*) ' col 1 on 5'
       write(lunspec,*) ' mark 1 on 2'
       write(lunspec,*) ' mark 1 on 3'
       write(lunspec,*) ' mark 1 on 4'
       write(lunspec,*) ' mark 1 on 5'
       write(lunspec,*) ' line off 2'
       write(lunspec,*) ' line off 3'
       write(lunspec,*) ' line off 4'
       write(lunspec,*) ' line off 5'
       write(lunspec,*) ' win 1'
       write(lunspec,*) ' yplot 1 2'
       write(lunspec,*) ' loc 0.05 0.45 0.57 1.0'
       write(lunspec,*) ' la 11 p 4.0 200.0 just left "Quad #1"'
       write(lunspec,*) ' r x1 0.15 15.0'
       write(lunspec,*) ' r y1 0.1 500.0'
       write(lunspec,*) ' la y Counts'
       write(lunspec,*) ' log x on'
       write(lunspec,*) ' log y on'
       write(lunspec,*) ' win 2'
       write(lunspec,*) ' yplot 1 3'
       write(lunspec,*) ' loc 0.47 0.45 0.99 1.0'
       write(lunspec,*) ' la 12 p 4.0 200.0 just left "Quad #2"'
       write(lunspec,*) ' r x2 0.15 15.0'
       write(lunspec,*) ' r y2 0.1 500.0'
       write(lunspec,*) ' log x on'
       write(lunspec,*) ' log y on'
       write(lunspec,*) ' win 3'
       write(lunspec,*) ' yplot 1 4'
       write(lunspec,*) ' loc 0.05 0.05 0.57 0.60'
       write(lunspec,*) ' la 13 p 4.0 200.0 just left "Quad #3"'
       write(lunspec,*) ' r x3 0.15 15.0'
       write(lunspec,*) ' r y3 0.1 500.0'
       write(lunspec,*) ' la y Counts'
       write(lunspec,*) ' la x Energy (keV)'
       write(lunspec,*) ' log x on'
       write(lunspec,*) ' log y on'
       write(lunspec,*) ' win 4'
       write(lunspec,*) ' yplot 1 5'
       write(lunspec,*) ' loc 0.47 0.05 0.99 0.60'
       write(lunspec,*) ' la 14 p 4.0 200.0 just left "Quad #4"'
       write(lunspec,*) ' r x4 0.15 15.0'
       write(lunspec,*) ' r y4 0.1 500.0'
       write(lunspec,*) ' col off 1'
       write(lunspec,*) ' la x Energy (keV)'
       write(lunspec,*) ' log x on'
       write(lunspec,*) ' log y on'
       !
       ! open the qdp file for writing
       !
       qdprate_F = "pn" // adjustl(trim(prefix)) // "-aug-rev-rate.qdp"
       ! create title for qdp plot
       call open_ascii_file(qdprate_F,lunrate)
       write(lunrate,*) ' viewport 0.2 0.2 0.9 0.8'
       write(lunrate,*) ' la t ', adjustl(trim(title))
       write(lunrate,*) ' read serr 2'
       write(lunrate,*) ' read serr 3'
       write(lunrate,*) ' read serr 4'
       write(lunrate,*) ' read serr 5'
       write(lunrate,*) ' read serr 6'
       write(lunrate,*) ' read serr 7'
       write(lunrate,*) ' read serr 8'
       write(lunrate,*) ' read serr 9'
       write(lunrate,*) ' font roman'
       write(lunrate,*) ' cs 1.5'
       write(lunrate,*) ' la f'
       write(lunrate,*) ' col 1 on 2'
       write(lunrate,*) ' col 2 on 3'
       write(lunrate,*) ' col 1 on 4'
       write(lunrate,*) ' col 3 on 5'
       write(lunrate,*) ' col 1 on 6'
       write(lunrate,*) ' col 4 on 7'
       write(lunrate,*) ' col 1 on 8'
       write(lunrate,*) ' col 5 on 9'
       write(lunrate,*) ' mark 1 on 2'
       write(lunrate,*) ' mark 1 on 3'
       write(lunrate,*) ' mark 1 on 4'
       write(lunrate,*) ' mark 1 on 5'
       write(lunrate,*) ' mark 1 on 6'
       write(lunrate,*) ' mark 1 on 7'
       write(lunrate,*) ' mark 1 on 8'
       write(lunrate,*) ' mark 1 on 9'
       write(lunrate,*) ' line off 2'
       write(lunrate,*) ' line off 3'
       write(lunrate,*) ' line off 4'
       write(lunrate,*) ' line off 5'
       write(lunrate,*) ' line off 6'
       write(lunrate,*) ' line off 7'
       write(lunrate,*) ' line off 8'
       write(lunrate,*) ' line off 9'
       write(lunrate,*) ' win 1'
       write(lunrate,*) ' yplot 1 2 3'
       write(lunrate,*) ' loc 0.05 0.45 0.57 1.0'
       write(lunrate,*) ' la 11 p 200.0 0.0036 just left "Quad #1"'
       write(lunrate,*) ' r x1 0.0 3500.0'
       write(lunrate,*) ' r y1 0.001 0.004'
       write(lunrate,*) ' la y Count Rate'
       write(lunrate,*) ' win 2'
       write(lunrate,*) ' yplot 1 4 5'
       write(lunrate,*) ' loc 0.47 0.45 0.99 1.0'
       write(lunrate,*) ' la 12 p 200.0 0.0036 just left "Quad #2"'
       write(lunrate,*) ' r x2 0.0 3500.0'
       write(lunrate,*) ' r y2 0.001 0.004'
       write(lunrate,*) ' win 3'
       write(lunrate,*) ' yplot 1 6 7'
       write(lunrate,*) ' loc 0.05 0.05 0.57 0.60'
       write(lunrate,*) ' la 13 p 200.0 0.0036 just left "Quad #3"'
       write(lunrate,*) ' r x3 0.0 3500.0'
       write(lunrate,*) ' r y3 0.001 0.004'
       write(lunrate,*) ' la y Count Rate'
       write(lunrate,*) ' la x Revolution'
       write(lunrate,*) ' win 4'
       write(lunrate,*) ' yplot 1 8 9'
       write(lunrate,*) ' loc 0.47 0.05 0.99 0.60'
       write(lunrate,*) ' la 14 p 200.0 0.0036 just left "Quad #4"'
       write(lunrate,*) ' r x4 0.0 3500.0'
       write(lunrate,*) ' r y4 0.001 0.004'
       write(lunrate,*) ' col off 1'
       write(lunrate,*) ' la x Revolution'
       !
       ! open the qdp file for writing
       !
       qdphard_F = "pn" // adjustl(trim(prefix)) // "-aug-rev-hard.qdp"
       ! create title for qdp plot
       call open_ascii_file(qdphard_F,lunhard)
       write(lunhard,*) ' viewport 0.2 0.2 0.9 0.8'
       write(lunhard,*) ' la t ', adjustl(trim(title))
       write(lunhard,*) ' read serr 2'
       write(lunhard,*) ' read serr 3'
       write(lunhard,*) ' read serr 4'
       write(lunhard,*) ' read serr 5'
       write(lunhard,*) ' read serr 6'
       write(lunhard,*) ' read serr 7'
       write(lunhard,*) ' read serr 8'
       write(lunhard,*) ' read serr 9'
       write(lunhard,*) ' font roman'
       write(lunhard,*) ' cs 1.5'
       write(lunhard,*) ' la f'
       write(lunhard,*) ' col 1 on 2'
       write(lunhard,*) ' col 2 on 3'
       write(lunhard,*) ' col 1 on 4'
       write(lunhard,*) ' col 3 on 5'
       write(lunhard,*) ' col 1 on 6'
       write(lunhard,*) ' col 4 on 7'
       write(lunhard,*) ' col 1 on 8'
       write(lunhard,*) ' col 5 on 9'
       write(lunhard,*) ' mark 1 on 2'
       write(lunhard,*) ' mark 1 on 3'
       write(lunhard,*) ' mark 1 on 4'
       write(lunhard,*) ' mark 1 on 5'
       write(lunhard,*) ' mark 1 on 6'
       write(lunhard,*) ' mark 1 on 7'
       write(lunhard,*) ' mark 1 on 8'
       write(lunhard,*) ' mark 1 on 9'
       write(lunhard,*) ' line off 2'
       write(lunhard,*) ' line off 3'
       write(lunhard,*) ' line off 4'
       write(lunhard,*) ' line off 5'
       write(lunhard,*) ' line off 6'
       write(lunhard,*) ' line off 7'
       write(lunhard,*) ' line off 8'
       write(lunhard,*) ' line off 9'
       write(lunhard,*) ' win 1'
       write(lunhard,*) ' yplot 1 2 3'
       write(lunhard,*) ' loc 0.05 0.45 0.57 1.0'
       write(lunhard,*) ' la 11 p 200.0 9.0 just left "Quad #1"'
       write(lunhard,*) ' r x1 0.0 3500.0'
       write(lunhard,*) ' r y1 0.0 10.0'
       write(lunhard,*) ' la y Hardness'
       write(lunhard,*) ' win 2'
       write(lunhard,*) ' yplot 1 4 5'
       write(lunhard,*) ' loc 0.47 0.45 0.99 1.0'
       write(lunhard,*) ' la 12 p 200.0 9.0 just left "Quad #2"'
       write(lunhard,*) ' r x2 0.0 3500.0'
       write(lunhard,*) ' r y2 0.0 10.0'
       write(lunhard,*) ' win 3'
       write(lunhard,*) ' yplot 1 6 7'
       write(lunhard,*) ' loc 0.05 0.05 0.57 0.60'
       write(lunhard,*) ' la 13 p 200.0 9.0 just left "Quad #3"'
       write(lunhard,*) ' r x3 0.0 3500.0'
       write(lunhard,*) ' r y3 0.0 10.0'
       write(lunhard,*) ' la y Hardness'
       write(lunhard,*) ' la x Revolution'
       write(lunhard,*) ' win 4'
       write(lunhard,*) ' yplot 1 8 9'
       write(lunhard,*) ' loc 0.47 0.05 0.99 0.60'
       write(lunhard,*) ' la 14 p 200.0 9.0 just left "Quad #4"'
       write(lunhard,*) ' r x4 0.0 3500.0'
       write(lunhard,*) ' r y4 0.0 10.0'
       write(lunhard,*) ' col off 1'
       write(lunhard,*) ' la x Revolution'
    endif
    !
    ! *** corner spectral file = old db_file ***
    ! next line opens the db_file fits file
    ! open/read the "CALDB" qpb fits file
    ! future: handled by CAL
    ! dies gracefully if file not found/unreadable/etc
    !
    if (setExists(trim(QPB_F)))then 
       QPB_S = dataSet(trim(QPB_F), READ) ! same as ftopen
    else
       write(errstr,*)&
            "QPB CAL file ",trim(QPB_F)," not found, exiting..."
       call error("noQPB",trim(errstr))
    end if
 !   QPB_T = table(QPB_S,"QPB")
 !   newRows = numberOfRows(QPB_T)
 !   ratePN = 1.447e-09
 !   hardPN = 1.61
 !   revlPN = 
    

    
    write(expression, '(a)') "FILTER!="//"'Closed'"
    print*,"Filter file with expression ", expression
  !  call message(AppMsg,Always)
  !  call initFilter( expression )       
  !  nRow1 = doFilter(inQPB_T)      ! Filter table
  !  print*,'old rows ',newRows,' new rows ',nRow1

    
    !
    !  Figure the correct extension name by chip number
    !
    write(extname,3010) quadN
3010 format('QUIES_QUAD',i1)
    write(ecode,3011) quadN
3011 format('noQUIES_QUAD',i1,'exten')
    
    if (.not.hasBlock(QPB_S,trim(extname))) then
       write(errstr,*)&
            "Input file does not contain extension : ",trim(extname)
       call error(trim(adjustl(ecode)),trim(errstr))
    else
       QPB_T = table(QPB_S,trim(extname)) ! read the whole extension
       qpbrows = numberOfRows(QPB_T)      ! get number of rows
    endif
    !
    !*** bhp: read selected colums in pn-qpb.fits file ***
    !*** Columns = KEY, REVOLUT, SUBMODE, EXPO, RATE, RATE_ERR
    !*** EXPO, HARD2, HARD2_ERR, SPECTRUM, SPECTRUM_OO
    !*** none of the others are actually used in calculations
    !
    ngr = 0
    nbr = 0
    if (.not.hasColumn(QPB_T,"KEY")) then
       call error("badColumn","Input file incorrect - no KEY column")
    else
       allocate (goodrows(qpbRows))
       goodrows = .FALSE.
       key_C = column(QPB_T,"KEY",READ) ! get "goodness of cleaning" key
       allocate(key(qpbRows))
       do i=1,qpbRows
          key(i) = stringCell(key_C,i-1)  ! fill chans array with keys
          if ((adjustl(trim(key(i))) == 'G') .or. &
               (adjustl(trim(key(i))) == '') .or. &
               (adjustl(trim(key(i))) == 'O')) then
             goodrows(i) = .TRUE. !keep track of good rows in QPB file
             ngr = ngr + 1
          else
             nbr = nbr + 1
          end if
       enddo
    endif
    !
    write(errstr,*) "First pass good key rows: ",ngr," bad key rows: ",nbr
    call message(AppMsg,Verbose)
    !
    if (ngr < 1) then 
       call error("noGoodQPB",&
            "Should not happend but found NO G/O/blank keys in QPB file!")
    end if
    if (.not.hasColumn(QPB_T,"REVOLUT")) then
       call error("badColumn","Input file incorrect - no REVOLUT column")
    else
       revl_C = column(QPB_T,"REVOLUT",READ)
       revl_ptr => int32Data(revl_C)
    endif
    if (.not.hasColumn(QPB_T,"SUB_MODE")) then ! stringCol
       call error("badColumn",&
            "Input file incorrect - no SUB_MODE column")
    else
       submode_C = column(QPB_T,"SUB_MODE",READ)
       allocate(submode(qpbRows))
       do i=i,qpbRows
          submode(i) = stringCell(submode_C,i)
       end do
    endif
    if (.not.hasColumn(QPB_T,"TOTAL_TIME")) then
       call error("badColumn",&
            "Input file incorrect - no TOTAL_TIME column")
    else
       ttime_C = column(QPB_T,"TOTAL_TIME",READ)
       ttime_ptr => Real64Data(ttime_C)
    endif
    if (.not.hasColumn(QPB_T,"EXPO")) then
       call error("badColumn","Input file incorrect - no EXPO column")
    else
       expo_C = column(QPB_T,"EXPO",READ)
       expo_ptr => Real64Data(expo_C)
    endif
    if (.not.hasColumn(QPB_T,"BACKSCAL")) then
       call error("badColumn","Input file incorrect - no BACKSCAL column")
    else
       back_C = column(QPB_T,"BACKSCAL",READ)
       back_ptr => int32Data(back_C)
    endif
    if (.not.hasColumn(QPB_T,"RATE")) then
       call error("badColumn","Input file incorrect - no RATE column")
    else
       rate_C = column(QPB_T,"RATE",READ)
       rate_ptr => Real64Data(rate_C)
    endif
    if (.not.hasColumn(QPB_T,"RATE_ERR")) then
       call error("badColumn","Input file incorrect - no RATE_ERR column")
    else
       urate_C = column(QPB_T,"RATE_ERR",READ)
       urate_ptr => Real64Data(urate_C)
    endif
    if (.not.hasColumn(QPB_T,"HARD_2")) then
       call error("badColumn","Input file incorrect - no HARD_2 column")
    else
       hard2_C = column(QPB_T,"HARD_2",READ)
       hard2_ptr => Real64Data(hard2_C)
    endif
    if (.not.hasColumn(QPB_T,"HARD_ERR_2")) then
       call error("badColumn","Input file incorrect - no HARD_ERR_2 column")
    else
       uhard2_C = column(QPB_T,"HARD_ERR_2",READ)
       uhard2_ptr => Real64Data(uhard2_C)
    endif
    if (.not.hasColumn(QPB_T,"SPECTRUM")) then
       call error("badColumn","Input file incorrect - no SPECTRUM column")
    else
       spec_C = column(QPB_T,"SPECTRUM",READ)
       spec_ptr => int32VectorData(spec_C) ! 4096J
    endif
    if (.not.hasColumn(QPB_T,"SPECTRUM_OO")) then
       call error("badColumn","Input file incorrect - no SPECTRUM_OO column")
    else ! dont these need to be 
       specOO_C = column(QPB_T,"SPECTRUM_OO",READ)
       specOO_ptr => int32VectorData(specOO_C) ! 4096J
    endif
    !
    ! now, get cols by looping through goodrows (from KEYS)
    !
    allocate(rate(ngr),urate(ngr),tspec(ngr),tspecoo(ngr),ootsca(ngr))
    allocate(hard2(ngr),uhard2(ngr),expo(ngr),ttime(ngr),revl(ngr))
    allocate(spec(chans,ngr),specoo(chans,ngr),back(ngr))
    !
    ngc = 1 ! counter for "good rows"
    do i=1,qpbRows
       if (goodrows(i)) then ! if key found 'good' above, aka .TRUE.
          revl(ngc) = revl_ptr(i-1)
          rate(ngc) = rate_ptr(i-1)
          back(ngc) = back_ptr(i-1)
          urate(ngc) = urate_ptr(i-1)
          hard2(ngc) = hard2_ptr(i-1)
          uhard2(ngc) = uhard2_ptr(i-1)
          tspec(1:chans) = spec_ptr(0:chans-1,i)
          tspecoo(1:chans) = specoo_ptr(0:chans-1,i)
          ttime(ngc) = ttime_ptr(i-1)
          expo(ngc) = expo_ptr(i-1)

          if (expo(ngc) > 100.) then ! do nothing if not?
             te1 = 0.
             do ii=plc_lo(1),plc_lo(2)
                te1 = te1 + tspec(ii)
             enddo
             te2 = 0.0
             do ii=plc_hi(1),plc_hi(2)
                te2 = te2 + tspec(ii)
             enddo
             do ii=1,chans
                spec(ii,ngc) = amscal*tspec(ii) &
                     /(expo(ngc)*back(ngc))
             enddo

             rate(ngc) = 1.0e6*rate(ngc)
             urate(ngc) = 1.0e6*urate(ngc)

             ! old if(ispecoo .ne. 0)

             teo1 = 0.0
             do ii=plc_lo(1),plc_lo(2)
                teo1 = teo1 + tspecoo(ii)
             enddo
             teo2 = 0.0
             do ii=plc_hi(1),plc_hi(2)
                teo2 = teo2 + tspecoo(ii)
             enddo
             do ii=1,chans
                if(ii < 600) then
                   temp = 5.0 - 1.3* &
                        (0.005*ii - 0.7)
                elseif(ii < 1400) then
                   temp = 2.0 - 0.125* &
                        (0.005*ii - 3.0)
                else
                   temp = 1.5
                endif
                specoo(ii,ngc) = amscal* &
                     tspecoo(ii)* &
                     temp/(expo(ngc)*back(ngc))
             enddo

             !  Set the OOT scale factor

             !  Set the PN oot scale factor, fix this for ALL submodes???!!!
!             if(trim(adjustl(submode(ngr))) .eq. 'PrimeFullWindow') then
!                ootsca = 0.0630
!             elseif(trim(adjustl(submode(ngr))) .eq. 'PrimeFullWindowExtended') then
!                if (frmtime .lt. 210) then
!                   ootsca = 0.0232
!                   smodefov = 'PrimeFullWindowExtended3'
!                else
!                   ootsca = 0.0163
!                   smodefov = 'PrimeFullWindowExtended5'
!                endif
!             else
!                call error("invalidSubmode",&
!                     "SUBMODE in FOV spectrum file not valid for this analysis at this time.")
!             endif
!             write(errstr,*)"PN submode/OOT Scale: ",trim(smodefov),ootsca
!             call message(AppMsg,Always)

             
             if(adjustl(trim(submode(ngc))) == &
                  'PrimeFullWindow') then
                tootsca = 0.063
             elseif(adjustl(trim(submode(ngc))) == &
                  'PrimeFullWindow') then
                tootsca = 0.063
             elseif(adjustl(trim(submode(ngc))) == &
                  'PrimeFullWindowExtended') then
                tootsca = 0.0232

                !  Full frame extended 5 is well hidden

             elseif(adjustl(trim(submode(ngc))) == &
                  'PrimeFullWindowExtended') then
                tootsca = 0.0163
             else
                write(errstr,*) "augment_pn: Bad submode: ",submode(ngc)
                call message(AppMsg,Always)
             end if
             ootsca(ngc) = tootsca
             !
             !
             !  Find the normal and OOT counts in the low and high
             !  spectral bands
             !
             te1 = 0.0
             teo1 = 0.0
             do ii=plc_lo(1),plc_lo(2)
                te1 = te1 + spec(ii,ngc)
                teo1 = teo1 + specoo(ii,ngc)
             enddo
             te2 = 0.0
             teo2 = 0.0
             do ii=plc_hi(1),plc_hi(2)
                te2 = te2 + spec(ii,ngc)
                teo2 = teo2 + specoo(ii,ngc)
             enddo
             temp = rate(ngc)
             rate(ngc) = te1 + te2 - &
                  tootsca*(teo1 + teo2)
             temp = hard2(ngc)
             hard2(ngc) = (te2 - tootsca*teo2)/ &
                  (te1 - tootsca*teo1)
          end if
          ngc = ngc + 1
       end if
       write(errstr,*) "augment_pn: Is ",i," a good row?: ",goodrows(i)
       call message(AppMsg,Noisy)
    end do
    write(errstr,*) "Second pass: good rows: ",ngr," rows above expo thresh: ",nbr
    call message(AppMsg,Verbose)

    !  Sort through the archived observation data collecting
    !  the specifics on possible observations that can be used
    !  for the augmentation.

    nobs = 0
    do i=1,ngc
       if((expo(i) > 0.0) .and. (ttime(i) > 0.0)) then
          if((expo(i) >= expmin) .and.  &
               (expo(i) < expmax) .and. &
               (rate(i) > ratemin) .and.  &
               (rate(i) < ratemax) .and.  &
               (hard2(i) > hardmin) .and. &
               (hard2(i) < hardmax) .and. &
               (expo(i) > 0.8*ttime(i))) then
             if (doplot) then
                select CASE (quadN)
                case (1)
                   write(lunhard,2820) revl(i), &
                        hard2(i), uhard2(i), d1, d2, &
                        d1, d2, d1, d2, &
                        d1, d2, d1, d2, &
                        d1, d2, d1, d2
2820               format(1x,i6,2f6.2,14f4.0)
                   write(lunrate,2720) revl(i), &
                        rate(i), urate(i), d1, d2, &
                        d1, d2, d1, d2, &
                        d1, d2, d1, d2, &
                        d1, d2, d1, d2
2720               format(1x,i6,2f10.6,14f4.0)
                   write(lunaug,2120) rate(i), urate(i), &
                        hard2(i), uhard2(i), d1, d2, &
                        d1, d2, d1, d2, &
                        d1, d2, d1, d2, &
                        d1, d2, d1, d2
2120               format(1x,2f10.6,2f6.2,22f4.0)
                case (2)
                   write(lunhard,2830) revl(i),  &
                        d1, d2, d1, d2,  &
                        hard2(i), uhard2(i), d1, d2, &
                        d1, d2, d1, d2, &
                        d1, d2, d1, d2
2830               format(1x,i6,4f4.0,2f6.2,10f4.0)
                   write(lunrate,2730) revl(i),  &
                        d1, d2, d1, d2,  &
                        rate(i), urate(i), d1, d2, &
                        d1, d2, d1, d2, &
                        d1, d2, d1, d2
2730               format(1x,i6,4f4.0,2f10.6,10f4.0)
                   write(lunaug,2130) rate(i), urate(i), &
                        d1, d2, d1, d2, &
                        hard2(i), uhard2(i), d1, d2, &
                        d1, d2, d1, d2, &
                        d1, d2, d1, d2
2130               format(1x,2f10.6,4f4.0,2f6.2,18f4.0)
                case (3)
                   write(lunhard,2840) revl(i),  &
                        d1, d2, d1, d2,  &
                        d1, d2, d1, d2, &
                        hard2(i), uhard2(i), d1, d2, &
                        d1, d2, d1, d2
2840               format(1x,i6,8f4.0,2f6.2,6f4.0)
                   write(lunrate,2740) revl(i),  &
                        d1, d2, d1, d2, &
                        d1, d2, d1, d2,  &
                        rate(i), urate(i), d1, d2, &
                        d1, d2, d1, d2
2740               format(1x,i6,8f4.0,2f10.6,6f4.0)
                   write(lunaug,2140) rate(i), urate(i), &
                        d1, d2, d1, d2, &
                        d1, d2, d1, d2, &
                        hard2(i), uhard2(i), d1, d2, &
                        d1, d2, d1, d2
2140               format(1x,2f10.6,8f4.0,2f6.2,14f4.0)
                case (4)
                   write(lunhard,2850) revl(i),  &
                        d1, d2, d1, d2,  &
                        d1, d2, d1, d2, &
                        d1, d2, d1, d2, &
                        hard2(i), uhard2(i), d1, d2
2850               format(1x,i6,12f4.0,2f6.2,2f4.0)
                   write(lunrate,2750) revl(i),  &
                        d1, d2, d1, d2, &
                        d1, d2, d1, d2,  &
                        d1, d2, d1, d2, &
                        rate(i), urate(i), d1, d2
2750               format(1x,i6,12f4.0,2f10.6,2f4.0)
                   write(lunaug,2150) rate(i), urate(i), &
                        d1, d2, d1, d2, &
                        d1, d2, d1, d2, &
                        d1, d2, d1, d2, &
                        hard2(i), uhard2(i), d1, d2
2150               format(1x,2f10.6,12f4.0,2f6.2,10f4.0)
                case default ! should NEVER happen! 
                   call error ("noQUAD",&
                        "augment_pn: quad is none of: [1-4]")
                end select
             end if
          else
             nb2 = nb2 + 1
          end if
       else
          nb3 = nb3 + 1
       end if
    end do
    !  Set the energies of the bins
    do i=1,chans
       energ(i) = 0.005*(i - 1)
    enddo

    !  Get some temporary values for the average rate and hardness
    !  errors for the corner data

    murate = 0.0
    muhard = 0.0
    do i=1,ngr
       murate = murate + urate(i)
       muhard = muhard + uhard2(i)
    enddo
    murate = murate/ngr
    muhard = muhard/ngr

    !  Find the closest c_tim secs

    if(c_tim == 0.0) then !should never happen, it's a declared parameter!!!
       call error ("noEXPlength",&
            "augment_pn: NO exposure length provided") 
    else
       if((c_dst(1) == 0.0) .and. (c_dst(2) == 0.0) .and. &
            (c_dst(3) ==  0.0)) then
          call error ("noDISTscal",&
               "augment_pn: No distance scalings")
       else

          !  First, set the distance in parameter space

          do i=1,ngr
             if(c_dst(1) /= 0.0) then
                rdist(i) = (rate(i)-ratein)/c_dst(1)
             else
                rdist(i) = 0.0
             endif
             if(c_dst(2) /= 0.0) then
                hdist(i) = (hard2(i)-hardin)/c_dst(2)
             else
                hdist(i) = 0.0
             endif
             if(c_dst(3) /= 0.0) then
                ddist(i) = (revl(i)-revlin)/c_dst(3)
             else
                ddist(i) = 0.0
             endif
             dist(i) = sqrt(rdist(i)**2 + hdist(i)**2 + &
                  ddist(i)**2)
          enddo
       endif
       !
       !  Sort through the CalDB data to order them by the distance
       !  from the observation values.
       !
       do i=1,ngr
          ord(i) = i
          tdist(i) = dist(i)
       enddo
       do ii=ngr-1,1,-1
          do i=1,ii
             if(tdist(i) > tdist(i+1)) then
                td = tdist(i)
                tdist(i) = tdist(i+1)
                tdist(i+1) = td
                itd = ord(i)
                ord(i) = ord(i+1)
                ord(i+1) = itd
             endif
          enddo
       enddo
       !
       !  Now accumulate spectra until a limiting exposure is reached
       !
       cum_exp = 0.0
       n = 0
       ave_rate = 0.0
       do while ((cum_exp < c_tim) .and. (n < ngr))
          n = n + 1
          cum_exp = cum_exp + expo(ord(n))
          ave_rate = ave_rate + rate(ord(n))*expo(ord(n))
          if(doplot) then
             select case (quadN)
             case (1)
                write(lunhard,2825) revl(ord(n)), &
                     d1, d2, hard2(ord(n)), uhard2(ord(n)), &
                     d1, d2, d1, d2, &
                     d1, d2, d1, d2, &
                     d1, d2, d1, d2
2825            format(1x,i6,2f4.0,2f6.2,12f4.0)
                write(lunrate,2725) revl(ord(n)), &
                     d1, d2, rate(ord(n)), urate(ord(n)), &
                     d1, d2, d1, d2, &
                     d1, d2, d1, d2, &
                     d1, d2, d1, d2
2725            format(1x,i6,2f4.0,2f10.6,12f4.0)
                write(lunaug,2125) rate(ord(n)), urate(ord(n)), &
                     d1, d2, hard2(ord(n)), uhard2(ord(n)), &
                     d1, d2, d1, d2, &
                     d1, d2, d1, d2, &
                     d1, d2, d1, d2
2125            format(1x,2f10.6,2f4.0,2f6.2,20f4.0)
             case (2)
                write(lunhard,2835) revl(ord(n)), &
                     d1, d2, d1, d2, &
                     d1, d2, hard2(ord(n)), uhard2(ord(n)), &
                     d1, d2, d1, d2, &
                     d1, d2, d1, d2
2835            format(1x,i6,6f4.0,2f6.2,8f4.0)
                write(lunrate,2735) revl(ord(n)), &
                     d1, d2, d1, d2, &
                     d1, d2, rate(ord(n)), urate(ord(n)), &
                     d1, d2, d1, d2, &
                     d1, d2, d1, d2
2735            format(1x,i6,6f4.0,2f10.6,8f4.0)
                write(lunaug,2135) rate(ord(n)), urate(ord(n)), &
                     d1, d2, d1, d2, &
                     d1, d2, hard2(ord(n)), uhard2(ord(n)), &
                     d1, d2, d1, d2, &
                     d1, d2, d1, d2
2135            format(1x,2f10.6,6f4.0,2f6.2,16f4.0)
             case (3)
                write(lunhard,2845) revl(ord(n)), &
                     d1, d2, d1, d2, &
                     d1, d2, d1, d2, &
                     d1, d2, hard2(ord(n)), uhard2(ord(n)), &
                     d1, d2, d1, d2
2845            format(1x,i6,10f4.0,2f6.2,4f4.0)
                write(lunrate,2745) revl(ord(n)), &
                     d1, d2, d1, d2, &
                     d1, d2, d1, d2, &
                     d1, d2, rate(ord(n)), urate(ord(n)), &
                     d1, d2, d1, d2
2745            format(1x,i6,10f4.0,2f10.6,4f4.0)
                write(lunaug,2145) rate(ord(n)), urate(ord(n)), &
                     d1, d2, d1, d2, &
                     d1, d2, d1, d2, &
                     d1, d2, hard2(ord(n)), uhard2(ord(n)), &
                     d1, d2, d1, d2
2145            format(1x,2f10.6,10f4.0,2f6.2,12f4.0)
             case (4)
                write(lunhard,2855) revl(ord(n)), &
                     d1, d2, d1, d2, &
                     d1, d2, d1, d2, &
                     d1, d2, d1, d2, &
                     d1, d2, hard2(ord(n)), uhard2(ord(n))
2855            format(1x,i6,14f4.0,2f6.2)
                write(lunrate,2755) revl(ord(n)), &
                     d1, d2, d1, d2, &
                     d1, d2, d1, d2, &
                     d1, d2, d1, d2, &
                     d1, d2, rate(ord(n)), urate(ord(n))
2755            format(1x,i6,14f4.0,2f10.6)
                write(lunaug,2155) rate(ord(n)), urate(ord(n)), &
                     d1, d2, d1, d2, &
                     d1, d2, d1, d2, &
                     d1, d2, d1, d2, &
                     d1, d2, hard2(ord(n)), uhard2(ord(n))
2155            format(1x,2f10.6,14f4.0,2f6.2,8f4.0)                   
             case default ! should NEVER happen! 
                call error ("noQUAD",&
                     "augment_pn: quad is none of: [1-4]")
             end select
          end if
       end do

       ave_rate = ave_rate/cum_exp
       write(errstr,4010) cum_exp
       call message(AppMsg,Always)
4010   format('Exposure: ',f10.1)

       write(errstr,4011) ave_rate
       call message(AppMsg,Always)
4011   format('Exposure weighted count rate: ',f10.6)

       write(errstr,4012) ratein
       call message(AppMsg,Always)
4012   format('Observation count rate:       ',f10.6)

       write(errstr,4013) hardin
       call message(AppMsg,Always)
4013   format('Observation hardness ratio:   ',f10.6)

       nb1 = ngr - n

       write(errstr,4014)&
            "Total/Good/Selected/NotSelected Obs: ",qpbRows,ngr,n,nb1
       call message(AppMsg,Always)
4014   format(A,4(1x,I4))

       write(errstr,4015)&
            "Out of Range/Bad Expo Obs: ",nb2,nb3
       call message(AppMsg,Always)
4015   format(A,2(1x,I4))

    endif
    !
    !  Actually put together the spectrum
    !
    aug_spec = 0.0
    aug_expo = 0. 
    do i=1,n
       do ii=1,chans
          aug_spec(ii) = aug_spec(ii) + (spec(ii,ord(i)) - &
               ootsca(ord(i))*specoo(ii,ord(i)))*expo(ord(i))
       enddo
       aug_expo = aug_expo + expo(ord(i))
    enddo
    do ii=1,chans
       aug_spec(ii) = aug_spec(ii)/aug_expo
    enddo
    te1 = 0.0
    te2 = 0.0
    do i=plc_lo(1),plc_lo(2)
       te1 = te1 + aug_spec(i)
    enddo
    do i=plc_hi(1),plc_hi(2)
       te2 = te2 + aug_spec(i)
    enddo

    write(errstr,4040) te1+te2
    call message(AppMsg,Always)
4040 format('Augmented count rate:     ',f10.6)
    write(errstr,4041) te1/te2
    call message(AppMsg,Always)
4041 format('Augmented hardness ratio: ',f8.3)
    write(errstr,4042) aug_expo
    call message(AppMsg,Always)
4042 format('Augmented exposure:       ',f10.1)

    if(doplot) then 
       do i=1,chans
          select case (quadN)
          case (1)
             write(lunspec,2225) energ(i), 1.0e6*aug_spec(i), &
                  d1, d1, d1
2225         format(1x,f10.3,f10.5,3f4.0)
          case (2)
             write(lunspec,2235) energ(i), d1, 1.0e6*aug_spec(i), &
                  d1, d1
2235         format(1x,f10.3,f4.0,f10.5,2f4.0)
          case (3)
             write(lunspec,2245) energ(i), d1, d1, &
                  1.0e6*aug_spec(i), d1
2245         format(1x,f10.3,2f4.0,f10.5,f4.0)
          case (4)
             write(lunspec,2255) energ(i), d1, d1, d1, &
                  1.0e6*aug_spec(i)
2255         format(1x,f10.3,3f4.0,f10.5)
          case default ! should NEVER happen! 
             call error ("noQUAD",&
                  "augment_pn: quad is none of: [1-4]")
          end select
       enddo
    endif

    call release (QPB_S)
    close(lunaug)
    close(lunspec)
    close(lunrate)
    close(lunhard)

    return

  end subroutine augment_pn
  !
  ! old call:
  !  subroutine spscale (xmmdir,xmmd,mosnum,maskfile,specfile,back,c)
  !
  subroutine spscale (mask_F,spec_F,backscal,c)
    !
    !  Date: 15 November 2006
    !  Conversion to F90 - 7 November 2012
    !  XMM-SAS CALDB compliance - BHP, 2016
    !
    !  Calculates the relative contribution of SP flares by region
    !
    !  spscale uses SP flare detector maps and region images to
    !  find the average SP flare contribution over the region.  It
    !  also returns the BACKSCALE value from the spectrum for the
    !  region.
    !  Input:
    !    QPB data directory
    !    MOS number
    !    Mask image file name
    !    Spectrum file name
    !  Output:
    !    BACKSCAL value for the region
    !    Average SP intensity for the region

    type(DataSetT) :: mask_S, spec_S, sprot_S
    type(TableT)   :: spec_T
    type(ArrayT)  :: mask_A, sprot_A

    character(len=1024),intent(in) :: mask_F, spec_F
    character(len=1024) :: sprot_F
    character(len=8) :: instrume, pre

    integer(kind=int16), dimension(:,:), pointer :: aDI16
    integer(kind=int32), dimension(:,:), pointer :: aDI32
    integer :: maskX, maskY, sprotX, sprotY, i, ii, n, status

    logical :: warndie = .TRUE.

    real :: backscal, c 
    real(kind=single), allocatable, dimension(:,:) :: mask, sprot !(780x780)
    real(kind=single), dimension(:,:), pointer :: aDR32
    real(kind=double), dimension(:,:), pointer :: aDR64

    ! Read in the mask image (e.g.: mos2S002-mask-im-400-750.fits)
    ! Mask file is a 780x780 I4 primary image array

    call check_open_set(trim(mask_F),mask_S,1,status,warndie)
    call read_2Darray(mask_S,mask_A,aDR32,aDR64,aDI16,aDI32,maskX,maskY)

    if (hasAttribute(mask_S,'INSTRUME')) then
       instrume = stringAttribute(mask_S,'INSTRUME')
       if (adjustl(trim(instrume)) == 'EPN') then
          pre = 'pn'
       elseif (adjustl(trim(instrume)) == 'EMOS1') then
          pre = 'mos1'
       elseif (adjustl(trim(instrume)) == 'EMOS2') then
          pre = 'mos2'
       else
          call error("notMOSorPN",& ! bhp: warn and default???
               "spscale: INSTRUME not EMOS1/EMOS2/EPN") !shouldn't happen!!
       endif
    else
       call error("maskNOinst",& ! bhp: warn and default???
            "spscale: input mask image has no INSTRUME")
    end if
    if ((maskX /= 780) .or. (maskY /= 780)) then
       write(errstr,*)"spscale: mask image not 780x780! X/Y = ",maskX,maskY
       call warning("not780x780",trim(errstr))
    endif
    allocate(mask(maskX,maskY))
    mask(1:maskX,1:maskY) = aDI32(0:maskX-1,0:maskY-1)

    ! Set the SP detector maps. These are CALDB files created
    ! by S.Snowden at GSFC. They contain soft proton maps in
    ! a 780x780 R4 primary image array.
    !
    ! bhp: sp-tot files should have INSTRUME set to EMOS1/EMOS2/EPN
    !      in future (not 'EPIC'), to sanity check vs mask file!!!
    !
    ! Get the 'correct' one by reading INSTRUME keyword from
    ! MASK file. Will be taken care of by CAL in future -- bhp

    sprot_F = adjustl(trim(pre)) // "-sp-tot.fits.gz"

    ! Read in the SP image (e.g.: mos2-sp-tot.fits.gz)

    call check_open_set(trim(sprot_F),sprot_S,1,status,warndie)
    call read_2Darray(sprot_S,sprot_A,aDR32,aDR64,aDI16,aDI32,sprotX,sprotY)
    if ((sprotX /= 780) .or. (sprotY /= 780)) then
       write(errstr,*)"spscale: sp image not 780x780! X/Y = ",sprotX,sprotY
       call warning("ImgNot780x780",trim(errstr))
    endif
    allocate(sprot(sprotX,sprotY))
    sprot(1:sprotX,1:sprotY) = aDR32(0:sprotX-1,0:sprotY-1)

    !
    ! read the mask spectrum FITS file
    !

    call check_open_set(trim(spec_F),spec_S,1,status,warndie)
    if (.not. hasBlock(spec_S,'SPECTRUM')) then  
       call error("noSPECTRUM",&
            "spscale: Input specfile has no SPECTRUM extension")
    end if
    spec_T = table(spec_S,'SPECTRUM', READ)

    if (.not.hasAttribute(spec_T,'BACKSCAL')) then  
       call error("noBACKSCAL",&
            "spscale: Input specfile file has no BACKSCAL keyword")
    end if
    backscal = real32attribute(spec_T,'BACKSCAL')    
    backscal = backscal / 1440000.0
    write(errstr,*) "spscale: backscale = ", backscal
    call message(AppMsg,Sparse)

    !  Loop over the image
    write(errstr,*) "spscale: looping over mask/image" 
    call message(AppMsg,Sparse)
    n = 0
    c = 0
    do i=1,maskX
       do ii=1,maskY
          if (mask(i,ii) > 0) then
             n = n + 1
             c = c + sprot(i,ii)
          endif
       enddo
    enddo
    c = c/n
    call release(mask_S)
    call release(spec_S)
    call release(sprot_S)

    return

  end subroutine spscale
  !
  !
  !
  subroutine hard_scale (xmmdir,alpha,mosnum,filt,islow,ishigh,fscal)
    !
    ! algorithm:
    !   1. open and read scale-factors CALDB files
    !      these are soft/med/hard band scale factors
    !      for normalizing between the 3 epic detectors
    !   2. calculate scale factor array
    !   3. apply array to create final scale factor to return
    !
    ! notes: called by comb and merge_comp_xmm
    !        xmmdir should be deprecated, scale factors
    !        CALDB files should be as outlined in bhp doc
    !        New ESAS CALDB file standard (2013)
    !

    type(DataSetT) :: sfs_S, sfm_S, sfh_S

    character(len=1024),intent(in) :: xmmdir
    character(len=8),intent(in) :: filt
    character(len=1024) :: sfs_F, sfm_F, sfh_F

    integer(kind=int16), dimension(:,:,:), pointer :: aDI16
    integer(kind=int32), dimension(:,:,:), pointer :: aDI32
    integer :: i, ii, iii, status
    integer :: sfsX,sfsY,sfsZ,sfmX,sfmY,sfmZ,sfhX,sfhY,sfhZ
    integer, intent(in) :: mosnum, islow, ishigh

    logical :: warndie = .TRUE.

    real(kind=single), intent(in) :: alpha
    real(kind=single), intent(out) :: fscal
    real(kind=single) :: ts, th
    real(kind=single), allocatable, dimension(:,:,:) :: &
         fscarr, fscarrs, fscarrm, fscarrh ! (240x240x9)
    real(kind=single), dimension(:,:,:), pointer :: aDR32
    real(kind=double), dimension(:,:,:), pointer :: aDR64    
    !
    ! call to CAL goes here here???:
    !
    sfs_F = adjustl(trim(xmmdir)) // "scale-factors-s.fits.gz"
    sfm_F = adjustl(trim(xmmdir)) // "scale-factors-m.fits.gz"
    sfh_F = adjustl(trim(xmmdir)) // "scale-factors-h.fits.gz"
    if((alpha < 1.0) .or. (alpha > 2.4)) then
       write(errstr,*)&
            "hard_scale: alpha out of range (1.0 > A > 2.4): ",alpha
       call error("badAlpha",trim(errstr))
    end if

    call check_open_set(trim(sfs_F),sfs_S,1,status,warndie)
    call read_3Darray(sfs_S,aDR32,aDR64,aDI16,aDI32,sfsX,sfsY,sfsZ)
    if ((sfsX /= 240) .or. (sfsY /= 240) .or. (sfsZ /= 240)) then
       write(errstr,*)&
            "hard_scale: scaleFactors image not 240x240x9! X/Y/Z = ",&
            sfsX,sfsy,sfsZ
       call error("ImgNot240x240x9",trim(errstr)) ! make a warning???
    endif
    allocate(fscarrs(sfsX,sfsY,sfsZ))
    fscarrs(1:sfsX,1:sfsY,1:sfsZ) = aDR32(0:sfsX-1,0:sfsY-1,0:sfsZ-1)

    call check_open_set(trim(sfm_F),sfm_S,1,status,warndie)
    call read_3Darray(sfm_S,aDR32,aDR64,aDI16,aDI32,sfmX,sfmY,sfmZ)
    if ((sfmX /= 240) .or. (sfmY /= 240) .or. (sfmZ /= 240)) then
       write(errstr,*)&
            "hard_scale: scaleFactors image not 240x240x9! X/Y/Z = ",&
            sfmX,sfmy,sfmZ
       call error("ImgNot240x240x9",trim(errstr))
    endif
    allocate(fscarrm(sfmX,sfmY,sfmZ))
    fscarrm(1:sfmX,1:sfmY,1:sfmZ) = aDR32(0:sfmX-1,0:sfmY-1,0:sfmZ-1)

    call check_open_set(trim(sfh_F),sfh_S,1,status,warndie)
    call read_3Darray(sfh_S,aDR32,aDR64,aDI16,aDI32,sfhX,sfhY,sfhZ)
    if ((sfhX /= 240) .or. (sfhY /= 240) .or. (sfhZ /= 240)) then
       write(errstr,*)&
            "hard_scale: scaleFactors image not 240x240x9! X/Y/Z = ",&
            sfhX,sfhy,sfhZ
       call error("ImgNot240x240x9",trim(errstr))
    endif
    allocate(fscarrh(sfhX,sfhY,sfhZ))
    fscarrh(1:sfhX,1:sfhY,1:sfhZ) = aDR32(0:sfhX-1,0:sfhY-1,0:sfhZ-1)
    !
    ! calculate combined scalefactor array
    ! bhp: replace with SELECT CASE fortran statement???
    !
    allocate(fscarr(sfsX,sfsY,sfsZ))

    if(alpha == 1.0) then
       do i=1,sfsX
          do ii=1,sfsY
             do iii=1,sfsZ
                fscarr(i,ii,iii) = fscarrh(i,ii,iii)
             enddo
          enddo
       enddo
    elseif(alpha == 1.7) then
       do i=1,sfsX
          do ii=1,sfsY
             do iii=1,sfsZ
                fscarr(i,ii,iii) = fscarrm(i,ii,iii)
             enddo
          enddo
       enddo
    elseif(alpha == 2.4) then
       do i=1,sfsX
          do ii=1,sfsY
             do iii=1,sfsZ
                fscarr(i,ii,iii) = fscarrs(i,ii,iii)
             enddo
          enddo
       enddo
    elseif(alpha < 1.7) then
       ts = (1.7 - alpha)/0.7
       th = (alpha - 1.0)/0.7
       do i=1,sfsX
          do ii=1,sfsY
             do iii=1,sfsZ
                fscarr(i,ii,iii) = &
                     ts*fscarrm(i,ii,iii) + &
                     th*fscarrh(i,ii,iii)
             enddo
          enddo
       enddo
    else
       ts = (2.4 - alpha)/0.7
       th = (alpha - 1.7)/0.7
       do i=1,sfsX
          do ii=1,sfsY
             do iii=1,sfsZ
                fscarr(i,ii,iii) = &
                     ts*fscarrs(i,ii,iii) + &
                     th*fscarrm(i,ii,iii)
             enddo
          enddo
       enddo
    endif

    !  Now evaluate the scale factor

    if(adjustl(trim(filt)) == 'Thin') then
       if(mosnum == 1) then
          fscal = fscarr(islow,ishigh,1)
       elseif(mosnum == 2) then
          fscal = fscarr(islow,ishigh,4)
       else
          fscal = fscarr(islow,ishigh,7)
       endif
    elseif(adjustl(trim(filt)) == 'Medi') then
       if(mosnum == 1) then
          fscal = fscarr(islow,ishigh,2)
       elseif(mosnum == 2) then
          fscal = 1.0
       else
          fscal = fscarr(islow,ishigh,8)
       endif
    elseif(adjustl(trim(filt)) == 'Thic') then
       if(mosnum == 1) then
          fscal = fscarr(islow,ishigh,3)
       elseif(mosnum == 2) then
          fscal = fscarr(islow,ishigh,6)
       else
          fscal = fscarr(islow,ishigh,9)
       endif
    else
       write(errstr,*)&
            "hard_scale: Filter is not Thin/Medi/Thic: ",trim(filt)
       call error("BadFilt",trim(errstr))

    endif

    return

    call release(sfs_S)
    call release(sfm_S)
    call release(sfh_S)

  end subroutine hard_scale


end module esas_mod
