! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
!
!+
! Module: gofobsolete_mod
!
! Authors:  Brendan Perry - GSFC 
!          
! Description: Numerical Recipes functions originally in ESAS package
!              but (seemingly) never used by any ESAS routines
! Comments: 
! Old f66 (!) and f77 (!) routines upgraded to f95 and SAS compliance
! 
!- 
module gofobsolete_mod

  use dal
  use errorHandling
  implicit none 

  public
contains  

  subroutine solve (ma,na,uu,yy,ssig,aa,cchisq,chicon,cvm)

    integer :: i, ii, j, ma, na
    real :: aa(na), cchisq, chicon(ma), ssig(ma), uu(ma,na), yy(ma)
    real(kind=double) chisq, chitmp, cvm(na,na), sum, thresh, tmp, wmax
    real(kind=double) :: tol = 1.0e-5
    real(kind=double), allocatable, dimension(:) :: a, b, sig, w, y
    real(kind=double), allocatable, dimension(:,:) :: u, v

    allocate(a(na))  
    allocate(b(ma)) 
    allocate(sig(ma))
    allocate(u(ma,na)) 
    allocate(v(na,na)) 
    allocate(w(na))  
    allocate(y(ma)) 

    write(errstr,*)"solve: n parameters = ",na
    call message(AppMsg,Noisy)
    write(errstr,*)"solve: n equations = ",ma
    call message(AppMsg,Noisy)

    do i=1,ma
       y(i) = yy(i)
       sig(i) = ssig(i)
       do ii=1,na
          u(i,ii) = uu(i,ii)
       enddo
    enddo

    do i=1,ma
       tmp = 1./sig(i)
       do ii=1,na
          u(i,ii) = u(i,ii)*tmp
       enddo
       b(i) = y(i)*tmp
    enddo

    call svdcmp(u,ma,na,w,v)

    wmax = 0.

    do j=1,na
       if(w(j) .gt. wmax) wmax = w(j)
    enddo

    thresh = tol*wmax

    do j=1,na
       if(w(j) .lt. thresh) w(j) = 0.
    enddo

    call svbksb(u,w,v,ma,na,b,a)

    call svdvar(v,na,w,cvm)

    ! calculate (uu*aa-yy)**2 for the fit as well as
    ! the contribution from each equation 

    chisq = 0.

    do i=1,ma
       sum = 0.
       do j=1,na
          sum = sum + a(j)*uu(i,j)
       enddo
       chitmp = ((y(i)-sum)/sig(i))**2
       if(chitmp .gt. 10.) then
          write(errstr,2000)i, chitmp, y(i), sig(i), sum
          call message(AppMsg,Always,trim(errstr))
2000      format(1x,i5,4f10.2)
       endif
       chisq = chisq + chitmp
       chicon(i) = chitmp
    enddo

    do i=1,na
       aa(i) = a(i)
    enddo
    cchisq = chisq

    return

  end subroutine solve

  subroutine svdcmp(a,m,n,w,v)

    integer :: i, its, j, k, l, m, mmax, n, nm
    real(kind=double) :: a(m,n), anorm, c, f, g, h, s, scale, &
         w(n), v(n,n), x, y, z
    real(kind=double) :: rv1(5000)

    g = 0.0
    scale = 0.0
    anorm = 0.0
    do i=1,n
       l = i + 1
       rv1(i) = scale*g
       g = 0.0
       s = 0.0
       scale = 0.0
       if (i.le.m) then
          do k=i,m
             scale = scale + dabs(a(k,i))
          enddo
          if (scale .ne. 0.0) then
             do k=i,m
                a(k,i) = a(k,i)/scale
                s = s + a(k,i)*a(k,i)
             enddo
             f = a(i,i)
             g = -dsign(dsqrt(s),f)
             h = f*g - s
             a(i,i) = f - g
             if (i .ne. n) then
                do j=l,n
                   s = 0.0
                   do k=i,m
                      s = s + a(k,i)*a(k,j)
                   enddo
                   f = s/h
                   do k=i,m
                      a(k,j) = a(k,j) + f*a(k,i)
                   enddo
                enddo
             endif
             do k= i,m
                a(k,i) = scale*a(k,i)
             enddo
          endif
       endif
       w(i) = scale*g
       g = 0.0
       s = 0.0
       scale = 0.0
       if ((i .le. m) .and. (i .ne. n)) then
          do k=l,n
             scale = scale + dabs(a(i,k))
          enddo
          if (scale .ne. 0.0) then
             do k=l,n
                a(i,k) = a(i,k)/scale
                s = s + a(i,k)*a(i,k)
             enddo
             f = a(i,l)
             g = -dsign(dsqrt(s),f)
             h = f*g - s
             a(i,l) = f - g
             do k=l,n
                rv1(k) = a(i,k)/h
             enddo

             if (i .ne. m) then
                do j=l,m
                   s = 0.0
                   do k=l,n
                      s = s + a(j,k)*a(i,k)
                   enddo
                   do k=l,n
                      a(j,k) = a(j,k) + s*rv1(k)
                   enddo
                enddo
             endif
             do k=l,n
                a(i,k) = scale*a(i,k)
             enddo
          endif
       endif
       anorm = max(anorm,(dabs(w(i))+dabs(rv1(i))))
    enddo

    do i=n,1,-1
       if (i .lt. n) then
          if (g .ne. 0.0) then
             do j=l,n
                v(j,i) = (a(i,j)/a(i,l))/g
             enddo
             do j=l,n
                s = 0.0
                do k=l,n
                   s = s + a(i,k)*v(k,j)
                enddo
                do k=l,n
                   v(k,j) = v(k,j) + s*v(k,i)
                enddo
             enddo
          endif
          do j=l,n
             v(i,j) = 0.0
             v(j,i) = 0.0
          enddo
       endif
       v(i,i) = 1.0
       g = rv1(i)
       l = i
    enddo

    do i=n,1,-1
       l = i + 1
       g = w(i)
       if (i.lt.n) then
          do j=l,n
             a(i,j) = 0.0
          enddo
       endif
       if (g .ne. 0.0) then
          g = 1.0/g
          if (i .ne. n) then
             do j=l,n
                s = 0.0
                do k=l,m
                   s = s + a(k,i)*a(k,j)
                enddo
                f = (s/a(i,i))*g
                do k=i,m
                   a(k,j) = a(k,j) + f*a(k,i)
                enddo
             enddo
          endif
          do j=i,m
             a(j,i) = a(j,i)*g
          enddo
       else
          do j= i,m
             a(j,i) = 0.0
          enddo
       endif
       a(i,i) = a(i,i) + 1.0
    enddo

    do k=n,1,-1
       do its=1,30
          do l=k,1,-1
             nm = l - 1
             if ((abs(rv1(l))+anorm).eq.anorm)  go to 2
             if ((abs(w(nm))+anorm).eq.anorm)  go to 1
          enddo
1         c = 0.0
          s = 1.0
          do i=l,k
             f = s*rv1(i)
             if ((abs(f)+anorm) .ne. anorm) then
                g = w(i)
                h = sqrt(f*f+g*g)
                w(i) = h
                h = 1.0/h
                c = (g*h)
                s = -(f*h)
                do j=1,m
                   y = a(j,nm)
                   z = a(j,i)
                   a(j,nm) = (y*c)+(z*s)
                   a(j,i) = -(y*s)+(z*c)
                enddo
             endif
          enddo
2         z = w(k)
          if (l .eq. k) then
             if (z .lt. 0.0) then
                w(k) = -z
                do j=1,n
                   v(j,k) = -v(j,k)
                enddo
             endif
             exit ! obsolete go to 3
          endif
          !                if (its.eq.30) pause 'no convergence in 30 iterations'
          x = w(l)
          nm = k-1
          y = w(nm)
          g = rv1(nm)
          h = rv1(k)
          f = ((y-z)*(y+z)+(g-h)*(g+h))/(2.0*h*y)
          g = sqrt(f*f+1.0)
          f = ((x-z)*(x+z)+h*((y/(f+sign(g,f)))-h))/x
          c = 1.0
          s = 1.0
          do j=l,nm
             i = j + 1
             g = rv1(i)
             y = w(i)
             h=s*g
             g=c*g
             z=sqrt(f*f+h*h)
             rv1(j)=z
             c=f/z
             s=h/z
             f= (x*c)+(g*s)
             g=-(x*s)+(g*c)
             h=y*s
             y=y*c
             do nm=1,n
                x=v(nm,j)
                z=v(nm,i)
                v(nm,j)= (x*c)+(z*s)
                v(nm,i)=-(x*s)+(z*c)
             enddo
             z=sqrt(f*f+h*h)
             w(j)=z
             if (z.ne.0.0) then
                z=1.0/z
                c=f*z
                s=h*z
             endif
             f= (c*g)+(s*y)
             x=-(s*g)+(c*y)
             do nm=1,m
                y=a(nm,j)
                z=a(nm,i)
                a(nm,j)= (y*c)+(z*s)
                a(nm,i)=-(y*s)+(z*c)
             enddo
          enddo
          rv1(l)=0.0
          rv1(k)=f
          w(k)=x
       enddo
       !3      continue
    enddo

    return
  end subroutine svdcmp
  !
  !
  !
  subroutine random (ran,idum)

    real :: r(97)
    real :: ran
    real :: rm1=3.8580247E-6, rm2=7.4373773E-6
    integer :: m1=259200, ia1=7141, ic1=54773
    integer :: m2=134456, ia2=8121, ic2=28411
    integer :: m3=243000, ia3=4561, ic3=51349
    integer :: iff=0
    integer :: idum, j
    integer :: ix1, ix2, ix3

    if ((idum .lt. 0) .or. (iff .eq. 0)) then
       iff = 1
       ix1 = mod(ic1 - idum,m1)
       ix1 = mod(ia1*ix1 + ic1,m1)
       ix2 = mod(ix1,m2)
       ix1 = mod(ia1*ix1 + ic1,m1)
       ix3 = mod(ix1,m3)
       do j=1,97
          ix1 = mod(ia1*ix1 + ic1,m1)
          ix2 = mod(ia2*ix2 + ic2,m2)
          r(j) = (float(ix1) + float(ix2)*rm2)*rm1
       enddo
       idum = 1
    endif

    ix1 = mod(ia1*ix1 + ic1,m1)
    ix2 = mod(ia2*ix2 + ic2,m2)
    ix3 = mod(ia3*ix3 + ic3,m3)

    j = 1 + (97*ix3)/m3
    if((j .gt. 97) .or. (j .lt. 1)) then
       write(errstr,*)"random: calculated J GT 97 or LT 1: ",j, "r = ",r
       call error("badJrandom",trim(errstr))
    endif
    ran = r(j)
    r(j) = (float(ix1) + float(ix2)*rm2)*rm1
    return
  end subroutine random

  subroutine gammlncal(gammln,xx)

    real gammln
    real(kind=double) :: cof(6)=(/76.18009173D0,-86.50532033D0,24.01409822D0, &
         -1.231739516D0,.120858003D-2,-.536382D-5/) 
    real(kind=double) :: fpf=5.5, half=0.5, one=1.0, ser, &
         stp=2.50662827465D0, tmp, x, xx
    integer :: j
    x = xx - one
    tmp = x + fpf
    tmp = (x + half)*log(tmp) - tmp
    ser = one
    do j=1,6
       x = x + one
       ser = ser + cof(j)/x
    enddo
    gammln = tmp + log(stp*ser)
    return

  end subroutine gammlncal
  !
  !
  !
  subroutine svdvar(v,n,w,cvm)

    integer :: i, j, k, n
    real(kind=double) :: cvm(n,n), v(n,n), w(n)
    real(kind=double) :: sum 
    real(kind=double), allocatable, dimension(:) :: wti

    allocate (wti(n))

    do i=1,n
       wti(i) = 0.
       if(w(i) .ne. 0.) wti(i)=1./(w(i)*w(i))
    enddo

    do i=1,n
       do j=1,i
          sum = 0.
          do k=1,n
             sum = sum + v(i,k)*v(j,k)*wti(k)
          enddo
          cvm(i,j) = sum
          cvm(j,i) = sum
       enddo
    enddo

    return
  end subroutine svdvar

  subroutine svbksb(u,w,v,m,n,b,x)

    integer i, j, jj, m, n
    real(kind=double) s, u(m,n), w(n), v(n,n), b(m), x(n)
    real(kind=double), allocatable, dimension(:) :: tmp

    allocate(tmp(m))

    do j=1,n
       s = 0.
       if(w(j) .ne. 0.)then
          do i=1,m
             s = s + u(i,j)*b(i)
          enddo
          s = s/w(j)
       endif
       tmp(j) = s
    enddo

    do j=1,n
       s = 0.
       do jj=1,n
          s = s + v(j,jj)*tmp(jj)
       enddo
       x(j) = s
    enddo

    return
  end subroutine svbksb


end module gofobsolete_mod
